import json
import os.path
import time

import click
import spotipy
from TheCodeLabs_BaseUtils.DefaultLogger import DefaultLogger
from spotipy import SpotifyOAuth, CacheFileHandler

from AudioRecorder import AudioRecorder
from Mp3Converter import Mp3Converter

LOG_FORMAT = '[%(levelname)-7s] - %(asctime)s - %(message)s'
LOGGER = DefaultLogger().create_logger_if_not_exists('SpotifyRecorder', logFormat=LOG_FORMAT)


class SpotifyRecorder:
    _MAX_WAIT_TIME_TRACK_STARTING_IN_S = 10
    _WAIT_TIME_TRACK_PLAYING_SHORT_IN_S = 1
    _THRESHOLD_TRACK_END_IN_MS = 5000

    def __init__(self, clientID: str,
                 clientSecret: str,
                 redirectUrl: str,
                 openBrowser: bool,
                 cacheFilePath: str,
                 playlists: list[dict[str, str | int]],
                 spotifyDeviceName: str,
                 audioDeviceName: str,
                 destinationFolder: str,
                 skipConfirmation: bool):
        self._clientID = clientID
        self._clientSecret = clientSecret
        self._redirectUrl = redirectUrl
        self._openBrowser = openBrowser
        self._cacheFilePath = cacheFilePath
        self._playlists = playlists
        self._spotifyDeviceName = spotifyDeviceName
        self._audioDeviceName = audioDeviceName
        self._destinationFolder = destinationFolder
        self._skipConfirmation = skipConfirmation

        os.makedirs(self._destinationFolder, exist_ok=True)

        self._spotify = self.login()

    def login(self) -> spotipy.Spotify:
        client_credentials_manager = SpotifyOAuth(client_id=self._clientID,
                                                  client_secret=self._clientSecret,
                                                  redirect_uri=self._redirectUrl,
                                                  scope='user-read-playback-state,user-modify-playback-state',
                                                  open_browser=self._openBrowser,
                                                  cache_handler=CacheFileHandler(cache_path=self._cacheFilePath))
        return spotipy.Spotify(client_credentials_manager=client_credentials_manager)

    def run(self):
        for playlist in self._playlists:
            self.__record_playlist(playlist)

    def __record_playlist(self, playlistSettings: dict[str, str | int]):
        LOGGER.info(f'Fetching all tracks for playlist "{playlistSettings["name"]}"...')
        startNumber = playlistSettings['startNumber']
        limit = playlistSettings['limit']
        if startNumber <= 0:
            raise ValueError(f'startNumber must be greater than 0')

        playlist = self.__get_playlist(playlistSettings['user'], playlistSettings['id'])
        tracks = self.__get_tracks(playlist)

        if limit == -1:
            LOGGER.info(f'Recording track #{startNumber} to end of playlist')
            tracks = tracks[startNumber - 1:]
        else:
            LOGGER.info(f'Recording track #{startNumber} to (including) #{startNumber + limit - 1}')
            tracks = tracks[startNumber - 1:startNumber - 1 + limit]

        totalDurationInSeconds = sum(
            [track['track']['duration_ms'] // 1000 for track in tracks if not track['is_local']])
        LOGGER.info(f'Total duration: {self.__convert_seconds_to_duration(totalDurationInSeconds)}')

        destinationFolder = os.path.join(self._destinationFolder, playlistSettings['name'])
        os.makedirs(destinationFolder, exist_ok=True)

        if self._skipConfirmation:
            self.__record_tracks(tracks, startNumber, destinationFolder)
        elif click.confirm('Do you want to start recording?', default=True):
            self.__record_tracks(tracks, startNumber, destinationFolder)
        else:
            LOGGER.warning('Aborted')

    def __get_playlist(self, username: str, playlistID: str) -> dict:
        LOGGER.info(f'Fetching playlist with ID: {playlistID} by {username}...')
        identifier = f'spotify:user:{username}:playlist:{playlistID}'
        playlist = self._spotify.playlist(identifier)
        LOGGER.info(f'Found playlist "{playlist["name"]}"')
        return playlist

    def __get_tracks(self, playlist) -> list[dict]:
        tracks = playlist['tracks']
        results = tracks['items']

        while tracks['next']:
            tracks = self._spotify.next(tracks)
            results.extend(tracks['items'])

        LOGGER.info(f'Found {len(results)} tracks in playlist')
        return results

    def __record_tracks(self, tracks: list, startNumber: int, destinationFolder: str):
        deviceId = self.__get_device_id_by_name(self._spotifyDeviceName)

        recordedTrackNumbers = []
        alreadyExistingTrackNumbers = []
        skippedTrackNumbers = []
        errorTrackNumbers = []

        for index, track in enumerate(tracks):
            indexInPlaylist = startNumber + index

            if track['is_local']:
                # It's not possible to add a local track to a playlist using the web API.
                # https://github.com/plamere/spotipy/issues/793#issuecomment-1082421408
                LOGGER.info(f'Skipping local track "{track["track"]["name"]}"')
                skippedTrackNumbers.append(indexInPlaylist)
                continue

            LOGGER.info(f'>>> Recording track {index + 1}/{len(tracks)}: '
                        f'#{indexInPlaylist} "{track["track"]["name"]}"...')
            try:
                recordingCreated = self.__record_single_track(deviceId, indexInPlaylist, track, destinationFolder)
                if recordingCreated:
                    recordedTrackNumbers.append(indexInPlaylist)
                else:
                    alreadyExistingTrackNumbers.append(indexInPlaylist)
            except Exception as e:
                LOGGER.error(f'An error occurred while recording track "{track["track"]["name"]}"', exc_info=e)
                errorTrackNumbers.append(indexInPlaylist)

        self.__print_end_statistics(errorTrackNumbers, recordedTrackNumbers, skippedTrackNumbers,
                                    alreadyExistingTrackNumbers, len(tracks))

    def __record_single_track(self, deviceId, indexInPlaylist, track, destinationFolder: str) -> bool:
        self.__stop_playback_if_playing(deviceId)

        fileName = self.__determine_file_name(indexInPlaylist, track, destinationFolder)
        filePathWav = f'{fileName}.wav'
        filePathMp3 = f'{fileName}.mp3'

        if os.path.exists(filePathMp3):
            LOGGER.info('Track already exists. Skipping.')
            return False

        recorder = AudioRecorder(self._audioDeviceName, filePathWav)

        with recorder.record():
            self.__play_track(deviceId, track['track']['uri'])
            timeWaitedForPlaying = self.__wait_for_track_playing(track['track']['id'])
            self.__wait_for_track_end(track, timeWaitedForPlaying)

        Mp3Converter.convert_to_mp3(filePathWav, filePathMp3)
        Mp3Converter.set_mp3_tags(filePath=filePathMp3,
                                  title=track['track']['name'],
                                  artist=self.__join_artists(track),
                                  album=track['track']['album']['name'])

        LOGGER.debug(f'\t\tRemoving wav file "{filePathWav}"')
        os.remove(filePathWav)

        return True

    @staticmethod
    def __print_end_statistics(errorTrackNumbers: list[int],
                               recordedTrackNumbers: list[int],
                               skippedTrackNumbers: list[int],
                               alreadyExistingTrackNumbers: list[int],
                               numberOfTracks: int):
        LOGGER.info('### DONE ###')

        LOGGER.info('>>> Skipped local tracks<<<')
        for number in skippedTrackNumbers:
            LOGGER.info(f'Skipped #{number}')

        LOGGER.info('>>> Errors <<<')
        for number in errorTrackNumbers:
            LOGGER.info(f'Error #{number}')

        LOGGER.info(f'### {numberOfTracks} tracks, '
                    f'{len(recordedTrackNumbers)} recorded, '
                    f'{len(skippedTrackNumbers)} local tracks skipped, '
                    f'{len(alreadyExistingTrackNumbers)} existing tracks skipped, '
                    f'{len(errorTrackNumbers)} errors ###')

    def __determine_file_name(self, index: int, track, destinationFolder: str) -> str:
        artists = self.__join_artists(track)
        fileName = f'{index} - {artists} - {track["track"]["name"]}'
        fileName = fileName.replace('/', '_')
        fileName = fileName.replace('\\', '_')
        fileName = fileName.replace('"', '')
        fileName = fileName.replace(':', '_')
        fileName = fileName.replace('*', '_')
        fileName = fileName.replace('?', '_')
        fileName = fileName.replace('>', '_')
        fileName = fileName.replace('<', '_')
        fileName = fileName.replace('|', '_')
        return os.path.join(destinationFolder, fileName)

    @staticmethod
    def __join_artists(track) -> str:
        artists = track['track']['artists']
        return ' & '.join(artist['name'] for artist in artists)

    def __wait_for_track_end(self, track, timeWaitedForPlaying: float) -> None:
        trackDurationInMs = track['track']['duration_ms']
        trackDurationInSeconds = trackDurationInMs // 1000
        LOGGER.info(f'\tTrack duration: {self.__convert_seconds_to_duration(trackDurationInSeconds)}')

        startTime = time.time()

        while time.time() - startTime < trackDurationInSeconds:
            currentPlayback = self._spotify.current_playback()
            if currentPlayback['is_playing']:
                remainingTimeInMs = trackDurationInMs - currentPlayback['progress_ms']
                remainingTimeInSeconds = remainingTimeInMs // 1000

                sleepTime = self._WAIT_TIME_TRACK_PLAYING_SHORT_IN_S

                if remainingTimeInMs > self._THRESHOLD_TRACK_END_IN_MS:
                    sleepTime = remainingTimeInSeconds / 2

                LOGGER.debug(f'\t\tWaiting for track to end (remaining: '
                             f'{self.__convert_seconds_to_duration(remainingTimeInSeconds)}, '
                             f'sleep: {self.__convert_seconds_to_duration(int(sleepTime))}s)...')
                time.sleep(sleepTime)
            else:
                waitDuration = int(time.time() - startTime)
                waitDuration += timeWaitedForPlaying

                if waitDuration < trackDurationInSeconds - self._WAIT_TIME_TRACK_PLAYING_SHORT_IN_S * 2:
                    raise RuntimeError(
                        f'Track finished too early (waited: {waitDuration}s, expected: {trackDurationInSeconds}s)')
                if waitDuration > trackDurationInSeconds + self._WAIT_TIME_TRACK_PLAYING_SHORT_IN_S * 3:
                    raise RuntimeError(
                        f'Track finished too late (waited: {waitDuration}s, expected: {trackDurationInSeconds})')

                LOGGER.info(f'Track finished. Waited {waitDuration}s, expected {trackDurationInSeconds}s, OK')
                break

    @staticmethod
    def __convert_seconds_to_duration(seconds: int):
        secs = seconds % 60
        minutes = int(seconds / 60) % 60
        hours = int(seconds / (60 * 60)) % 24
        return f'{str(hours).zfill(2)}:{str(minutes).zfill(2)}:{str(secs).zfill(2)}'

    def __get_device_id_by_name(self, deviceName: str) -> str:
        devices = self._spotify.devices()['devices']
        for device in devices:
            if device['name'] == deviceName:
                return device['id']

        raise RuntimeError('Device not found')

    def __play_track(self, deviceId: str, trackUri: str) -> None:
        self._spotify.start_playback(device_id=deviceId, uris=[trackUri])

    def __stop_playback_if_playing(self, deviceId: str) -> None:
        currentPlayback = self._spotify.current_playback()
        if currentPlayback is None:
            return

        if currentPlayback['is_playing']:
            self._spotify.pause_playback(device_id=deviceId)

    def __wait_for_track_playing(self, expectedTrackId: str) -> float:
        LOGGER.debug(f'\t\tWait for track to start playing...')
        startTime = time.time()
        duration = 0
        while time.time() - startTime < self._MAX_WAIT_TIME_TRACK_STARTING_IN_S:
            currentPlayback = self._spotify.current_playback()
            if currentPlayback['is_playing']:
                duration = time.time() - startTime
                if currentPlayback['item']['id'] == expectedTrackId:
                    LOGGER.debug(f'\t\tTrack started playing after {duration:.1f}s')
                    break
                else:
                    raise RuntimeError(
                        f'Wrong track started playing (actual: {currentPlayback["item"]["id"]}, expected: {expectedTrackId})')

            time.sleep(1)

        return duration


if __name__ == '__main__':
    with open('config/settings-recorder.json', 'r', encoding='utf-8') as f:
        SETTINGS = json.load(f)

    spotifyBackup = SpotifyRecorder(SETTINGS['spotifyAPI']['clientID'],
                                    SETTINGS['spotifyAPI']['clientSecret'],
                                    SETTINGS['redirectUrl'],
                                    SETTINGS['openBrowser'],
                                    SETTINGS['cacheFilePath'],
                                    SETTINGS['playlists'],
                                    SETTINGS['spotifyDeviceName'],
                                    SETTINGS['audioDeviceName'],
                                    SETTINGS['destinationFolder'],
                                    SETTINGS['skipConfirmation'],
                                    )

    spotifyBackup.run()

    LOGGER.info('### DONE ###')
